from PIL.ImageDraw import Draw


class Tree:
    """
    Tree class to storeing a graphical representation of glycans in SNFG form.
    """

    def __init__(self, glycan):
        """
        Initialize the tree with the glycan.

        Args:
            glycan (Glycan): Glycan object to be represented in this tree object
        """
        self.root = Node(0, 0, glycan.parse_tree)

    def assign_coords(self):
        """
        Assign coordinates to every node in this tree.

        Returns:
            It-self
        """
        self.root.assign_coords()
        return self

    def adjust_coords(self, offset):
        """
        Adjust the coordinates by adjusting every node's coordinate.

        Args:
            offset (float): Offset by which all nodes should be moved
        """
        self.root.adjust_coords(offset)

    def get_bounds(self):
        """
        Bounding box in terms of node coordinates on a graphical representation.

        Returns:
            Tuple of 4 floats representing upper left and bottom right corners
        """
        return 0, 0, self.root.get_max_depth(), self.root.get_lower_bound()[-1][1]

    def draw_edges(self, img, **params):
        """
        Draw the edges of the tree into the ImageDraw object.

        Args:
            img (Draw): ImageDraw object to draw on
            **params: parameters for visualization
        """
        self.root.draw_edges(img, **params)

    def draw_nodes(self, img, glycan, **params):
        """
        Draw the nodes of the tree representing the individual monosaccharides in a sugar polymer.

        Args:
            img (Draw): ImageDraw object to draw on
            glycan (Glycan): Glycan object holding some information on the sugar monomers
            **params: parameters for visualization
        """
        self.root.draw_nodes(img, glycan.parse_tree, **params)


class Node:
    """
    A single node in representing a monosaccharide in a graphical SNFG representation.
    """
    def __init__(self, idx, depth, parse_tree):
        """
        Initialize a node representing a monosaccharide.

        Args:
            idx (int): Id of the glycan that matches the id in the networkx tree of the glycan object in the parse_tree.
            depth (int): Depth of the node in the tree, root has depth 0, leaves have the highest depth-value
            parse_tree (networkx.DiGraph): networkx graph representation of the glycan computed while parsing
        """
        self.idx = idx
        self.depth = depth
        self.children = []
        self.coord = None

        # iterate over the children and recursively initialize them.
        for _, cidx in parse_tree.edges(self.idx):
            self.children.append((
                Node(cidx, self.depth + 1, parse_tree),
                parse_tree.get_edge_data(self.idx, cidx)["type"]
            ))

        # sort the children based on the depth of their most far away leaf and the number of children
        self.children = list(
            sorted(
                self.children,
                key=lambda child: (child[0].get_max_depth(), child[0].get_num_children()),
                reverse=True
            )
        )

    def assign_coords(self):
        """
        Assign coordinates of this node. It is only necessary to assign y-coordinates as x coordinates are set to depth.

        Returns:
            The own y-coordinate relative to the parent.
        """
        # initially the coordinate is 0 and can be returned in case this node is a leaf
        self.coord = 0
        if len(self.children) == 0:
            return 0

        # compute the coordinates of the children
        for child, _ in self.children:
            child.assign_coords()

        # iterate over the children and arrange them to not overlap by moving the lower child down based on bounds
        for i in range(1, len(self.children)):
            lower = self.children[i - 1][0].get_lower_bound()
            upper = self.children[i][0].get_upper_bound()
            offset = max(l - u for (_, l), (_, u) in zip(lower, upper)) + 1
            self.children[i][0].adjust_coords(offset)

        # set own coordinate to be in middle height of their children
        self.coord = sum(child.coord for child, _ in self.children) / len(self.children)

        return self.coord

    def adjust_coords(self, offset):
        """
        Adjust the y-coordinates of a node by adding the offset to the coordinate and propagating the offset to the
        children.

        Args:
            offset (float): Offset to be added to the subtrees coordinates
        """
        self.coord += offset
        for child, _ in self.children:
            child.adjust_coords(offset)

    def get_lower_bound(self):
        """
        Get the lower bound of this subtree in terms of node coordinates.

        Returns:
            A List of tuples of x and y coordinates (floats) representing the lower bound for this subtree.
        """
        # compose the lower bound as the own coordinate the lower bound of the lowest child
        if len(self.children) != 0:
            return [(self.depth, self.coord)] + self.children[-1][0].get_lower_bound()
        return [(self.depth, self.coord)]

    def get_upper_bound(self):
        """
        Get the upper bound of this subtree in terms of node coordinates.

        Returns:
            A List of tuples of x and y coordinates (floats) representing the upper bound for this subtree.
        """
        # compose the upper bound as the own coordinate the upper bound of the highest child
        if len(self.children) != 0:
            return [(self.depth, self.coord)] + self.children[0][0].get_upper_bound()
        return [(self.depth, self.coord)]

    def get_num_children(self):
        """
        Get the number of children of this node.

        Returns:
            Number of children
        """
        return len(self.children)

    def get_max_depth(self):
        """
        Get the maximal depth value of a leaf of the subtree of this node. The return value will be relative to the
        root not to this node.

        Returns:
            Depth of the deepest leaf of this subtree
        """
        if len(self.children) == 0:
            return self.depth
        return max(child.get_max_depth() for child, _ in self.children)

    def get_coords(self):
        """
        Get the coordinates pair of this node.

        Returns:
            Tuple of two floats representing the x and y coordinate of this node
        """
        return self.depth, self.coord

    def draw_edges(self, img, **params):
        """
        Draw the edge between two nodes.

        Args:
            img (Draw): ImageDraw object to draw on
            **params: parameters for visualization
        """
        x, y = self.get_coords()
        k = 0.3 * (params["box"] / params["width"])
        l = 0.3 * (params["box"] / params["width"])
        for child, info in self.children:
            # draw a line from the center of this node to the centers of the child nodes
            c_x, c_y = child.get_coords()
            img.line([
                ((x + 0.5) * params["width"], (y + 0.5) * params["height"]),
                ((c_x + 0.5) * params["width"], (c_y + 0.5) * params["height"])
            ], fill="black", width=params["line"])
            child.draw_edges(img, **params)

            # for each edge denote the additional information such as a/b conformation and binding carbon id
            img.text(
                ((x + k + 0.5) * params["width"], (c_y + (y - c_y) * (1 - l) + 0.38) * params["height"]),
                text=info[1],
                fill="black",
                anchor="lb",
            )
            img.text(
                ((x - k + 1.48) * params["width"], (c_y + (y - c_y) * l + 0.38) * params["height"]),
                text=info[-2],
                fill="black",
                anchor="lb",
            )

    def draw_nodes(self, img, tree, **params):
        """
        Draw the SNFG representation of this monomer to the ImageDraw object.

        Args:
            img (Draw): ImageDraw object to draw on
            tree (networkx.DiGraph): ParseTree from a glycan, computed during parsing of the IUPAC string
            **params: parameters for visualization
        """
        # draw this node and all the children recursively
        draw_glycan(img, *(self.get_coords()), *snfg_fy(tree.nodes[self.idx]["type"]), **params)
        for child, _ in self.children:
            child.draw_nodes(img, tree, **params)


colors = {
    "Hexose": {  # same for Hexose, HexNAc, Hexosamine, Hexuronate, Unknown
        "Glc": "blue",
        "Man": "green",
        "Gal": "yellow",
        "Gul": "orange",
        "Alt": "lightpink",
        "All": "BlueViolet",
        "Tal": "LightSkyBlue",
        "Ido": "sienna",  # brown
    },
    "Deoxyhexose": {
        "Qui": "blue",
        "Rha": "green",
        "6dGul": "orange",
        "6dAlt": "lightpink",
        "6dTal": "LightSkyBlue",
        "Fuc": "red",  # brown
    },
    "DeoxyhexNAc": {
        "QuiNAc": "blue",
        "RhaNAc": "green",
        "6dAltNAc": "lightpink",
        "6dTalNAc": "LightSkyBlue",
        "FucNAc": "red",  # brown
    },
    "Di-deoxyhexose": {
        "Oli": "blue",
        "Tyv": "green",
        "Abe": "orange",
        "Par": "lightpink",
        "Dig": "BlueViolet",
        "Col": "LightSkyBlue",
    },
    "Pentose": {
        "Ara": "green",
        "Lyx": "yellow",
        "Xyl": "orange",
        "Rib": "lightpink",
    },
    "3-dideoxy-nunolusonic acids": {
        "Kdn": "green",
        "Neu5Ac": "BlueViolet",
        "neu5Gc": "LightSkyBlue",
        "Neu": "sienna",  # brown
        "Sia": "red",
    },
    "3,9-dideoxy-nunolusonic acids": {
        "Pse": "green",
        "Leg": "yellow",
        "Aci": "lightpink",
        "4eLeg": "LightSkyBlue",
    },
    "Unknown": {  # same for Hexose, HexNAc, Hexosamine, Hexuronate, Unknown
        "Bac": "blue",
        "LDmanHep": "green",
        "Kdo": "yellow",
        "Dha": "orange",
        "DDmanHep": "lightpink",
        "MurNAc": "BlueViolet",
        "MurNGc": "LightSkyBlue",
        "Mur": "sienna",  # brown
    },
    "Assigned": {  # same for Hexose, HexNAc, Hexosamine, Hexuronate, Unknown
        "Api": "blue",
        "Fru": "green",
        "Tag": "yellow",
        "Sor": "orange",
        "Psi": "lightpink",
    },
}


classes = {
    "Hexose": set(colors["Hexose"].keys()),
    "HexNAc": set([x + "NAc" for x in colors["Hexose"].keys()]),
    "Hexosamine": set([x + "N" for x in colors["Hexose"].keys()]),
    "Hexuronate": set([x + "A" for x in colors["Hexose"].keys()]),
    "Deoxyhexose": {"Qui", "Rha", "6dGul", "6dAlt", "6dTal", "Fuc"},
    "DeoxyhexNAc": {"QuiNAc", "RhaNAc", "6dAltNAc", "6dTalNAc", "FucNAc"},
    "Di-deoxyhexose": {"Oli", "Tyv", "Abe", "Par", "Dig", "Col"},
    "Pentose": {"Ara", "Lyx", "Xyl", "Rib"},
    "3-deoxy-nonulosonic acids": {"Kdn", "Neu5Ac", "Neu5Gc", "Neu", "Sia"},
    "3,9-dideoxy-nonulosonic acids": {"Pse", "Leg", "Aci", "4eLeg"},
    "Unknown": {"Bac", "LDmanHep", "Kdo", "Dha", "DDmanHep", "MurNAc", "MurNGc", "Mur"},
    "Assigned": {"Api", "Fru", "Tag", "Sor", "Psi"},
}


def snfg_fy(glycan):
    """
    Convert the description of a glycan into the SNFG name and a list of attached functional groups.

    Args:
        glycan (Glycan): Glycan holding the information to be extracted

    Returns:
        Name of the glycan in SNFG-conform term and a list of remaining functional groups.
    """
    name = glycan.get_name()
    names = set(list(zip(*(glycan.get_recipe())))[0])
    if name in {
        "Oli", "Tyv", "Abe", "Par", "Dig", "Col", "Ara", "Lyx", "Xyl", "Rib", "Kdn", "Sia", "Pse", "Aci",
        "Bac", "Kdo", "Dha", "Api", "Fru", "Tag", "Sor", "Psi"
    }:
        return name, names.difference({name})
    if name in classes["Hexose"] and ("NAc" in names or "N" in names or "A" in names):
        if "NAc" in names and not (name in {"Alt", "Tal"} and "6d" in names):
            return name + "NAc", names.difference({"NAc", name})
        if "N" in names:
            return name + "N", names.difference({"N", name})
        if "A" in names:
            return name + "A", names.difference({"A", name})
    if name == "Gul" and "6d" in names:
        return "6dGul", names.difference({"6d", name})
    if name in {"Alt", "Tal"} and ("NAc" in names or "6d" in names):
        if "NAc" in names and "6d" in names:
            return "6d" + name + "NAc", names.difference({"6d", "NAc", name})
        if "NAc" in names:
            return name + "NAc", names.difference({"NAc", name})
        if "6d" in names:
            return "6d" + name, names.difference({"6d", name})
    if name in {"Qui", "Rha", "Fuc", "Mur"} and "NAc" in names:
        return name + "NAc", names.difference({"NAc", name})
    if name == "Neu" and "5Ac" in names or "Ac" in names:
        return "Neu5Ac", names.difference({"5Ac", name})
    if name == "Neu" and "5Gc" in names or "Gc" in names:
        return "Neu5Gc", names.difference({"5Gc", name})
    if name == "Mur" and "NGc" in names:
        return "Mur5Gc", names.difference({"5Gc", name})
    if name == "Leg" and "4e" in names:
        return "4eLeg", names.difference({"4e", name})
    if name == "Man" and "LD" in names and "Hep" in names:
        return "LDManHep", names.difference({"LD", "Man", "Hep"})
    if name == "Man" and "DD" in names and "Hep" in names:
        return "DDManHep", names.difference({"DD", "Man", "Hep"})
    return name, names.difference({name})


def draw_glycan(img, x, y, name, groups, **params):
    """
    Draw a single glycan node to the image.

    Args:
        img (Draw): ImageDraw object to paint on
        x (float): x coordinate to be drawn
        y (float): y coordinate to be drawn
        name (str): name of the monomer to be drawn
        groups (List[str]): List of strings to be written beyond the
        **params: See description in viz.create_snfg_image
    """
    if name in classes["Hexose"]:
        draw_hexose(img, x, y, name, **params)
    elif name in classes["HexNAc"]:
        draw_hexnac(img, x, y, name, **params)
    elif name in classes["Hexosamine"]:
        draw_hexosamine(img, x, y, name, **params)
    elif name in classes["Hexuronate"]:
        draw_hexuronate(img, x, y, name, **params)
    elif name in classes["Deoxyhexose"]:
        draw_deoxyhexose(img, x, y, name, **params)
    elif name in classes["DeoxyhexNAc"]:
        draw_deoxyhexnac(img, x, y, name, **params)
    elif name in classes["Di-deoxyhexose"]:
        draw_di_deoxyhexose(img, x, y, name, **params)
    elif name in classes["Pentose"]:
        draw_pentose(img, x, y, name, **params)
    elif name in classes["3-deoxy-nonulosonic acids"]:
        draw_3_deoxy_nonulosonic_acids(img, x, y, name, **params)
    elif name in classes["3,9-dideoxy-nonulosonic acids"]:
        draw_3_9_dideocy_nonulosonic_acids(img, x, y, name, **params)
    elif name in classes["Unknown"]:
        draw_unknown(img, x, y, name, **params)
    elif name in classes["Assigned"]:
        draw_assigned(img, x, y, name, **params)
    else:
        draw_other(img, x, y, **params)

    # finally write the text under the symbol
    text, width = arrange_node_groups(groups, img, **params)
    img.multiline_text(
        ((x + 0.5) * params["width"] - width * 0.5, (y + 0.8) * params["height"]),
        text,
        fill="black",
        anchor="ma",
        align="center"
    )


def arrange_node_groups(groups, img, **params):
    """
    Arrange the names of the attached functional groups in a way to not exceed the width of the sign.

    Args:
        groups (List[str]): List of descriptions of functional groups attached to a glycan
        img (Draw): ImageDraw object to estimate the width of each textbox
        **params: See description in Glycan.create_snfg_image

    Returns:
        Text with newlines and maximal width of the resulting textbox for better centering it.
    """
    # set some variable and define a threshold for the maximal width of a line
    output = ""
    length = 0
    threshold = 0.5 * params["box"]
    lines = []

    # iterate over all groups
    for g in groups:
        # if the line is empty, add the functional group as new line and store the length in case it's below threshold
        if length == 0:
            output += g
            tl = img.textlength(g)
            if tl > threshold:
                output += "\n"
                lines.append(tl)
            else:
                length = tl
        # otherwise, check if the line has enough space for another functional group and add a new line eventually
        else:
            tl = img.textlength(g)
            if length + tl > threshold:
                output += "\n" + g
                lines.append(length)
                length = tl
            else:
                output += g
                length += tl
    lines.append(length)

    return output, max(lines)


def draw_hexose(img, x, y, name, **params):
    """
    Draw a filled circle for an assigned glycan.

    Args:
        img (Draw): ImageDraw object to use for drawing.
        x (float): x coordinate of the shape on the image before offsetting
        y (float): y coordinate of the shape on the image before offsetting
        name (str): name of the glycan to select the correct color
        **params: See description in Glycan.create_snfg_image
    """
    img.ellipse([
        ((x + 0.5) * params["width"] - 0.25 * params["box"], (y + 0.5) * params["height"] - 0.25 * params["box"]),
        ((x + 0.5) * params["width"] + 0.25 * params["box"], (y + 0.5) * params["height"] + 0.25 * params["box"])
    ], fill=colors["Hexose"][name], width=params["stroke"], outline="black")


def draw_hexnac(img, x, y, name, **params):
    """
    Draw a filled square for an assigned glycan.

    Args:
        img (Draw): ImageDraw object to use for drawing.
        x (float): x coordinate of the shape on the image before offsetting
        y (float): y coordinate of the shape on the image before offsetting
        name (str): name of the glycan to select the correct color
        **params: See description in Glycan.create_snfg_image
    """
    img.rectangle([
        ((x + 0.5) * params["width"] - 0.25 * params["box"], (y + 0.5) * params["height"] - 0.25 * params["box"]),
        ((x + 0.5) * params["width"] + 0.25 * params["box"], (y + 0.5) * params["height"] + 0.25 * params["box"]),
    ], fill=colors["Hexose"][name[:3]], width=params["stroke"], outline="black")


def draw_hexosamine(img, x, y, name, **params):
    """
    Draw a crossed square for an assigned glycan.

    Args:
        img (Draw): ImageDraw object to use for drawing.
        x (float): x coordinate of the shape on the image before offsetting
        y (float): y coordinate of the shape on the image before offsetting
        name (str): name of the glycan to select the correct color
        **params: See description in Glycan.create_snfg_image
    """
    img.polygon([
        ((x + 0.5) * params["width"] - 0.25 * params["box"], (y + 0.5) * params["height"] - 0.25 * params["box"]),
        ((x + 0.5) * params["width"] + 0.25 * params["box"], (y + 0.5) * params["height"] + 0.25 * params["box"]),
        ((x + 0.5) * params["width"] + 0.25 * params["box"], (y + 0.5) * params["height"] - 0.25 * params["box"]),
    ], fill=colors["Hexose"][name[:3]], width=params["stroke"], outline="black")
    img.polygon([
        ((x + 0.5) * params["width"] - 0.25 * params["box"], (y + 0.5) * params["height"] - 0.25 * params["box"]),
        ((x + 0.5) * params["width"] + 0.25 * params["box"], (y + 0.5) * params["height"] + 0.25 * params["box"]),
        ((x + 0.5) * params["width"] - 0.25 * params["box"], (y + 0.5) * params["height"] + 0.25 * params["box"]),
    ], fill="white", width=params["stroke"], outline="black")


def draw_hexuronate(img, x, y, name, **params):
    """
    Draw a divided diamond for an assigned glycan.

    Args:
        img (Draw): ImageDraw object to use for drawing.
        x (float): x coordinate of the shape on the image before offsetting
        y (float): y coordinate of the shape on the image before offsetting
        name (str): name of the glycan to select the correct color
        **params: See description in Glycan.create_snfg_image
    """
    if name in {"IdoA", "AltA"}:
        cols = ("white", colors["Hexose"][name[:3]])
    else:
        cols = (colors["Hexose"][name[:3]], "white")
    img.polygon([
        ((x + 0.5) * params["width"] - 0.25 * params["box"], (y + 0.5) * params["height"]),
        ((x + 0.5) * params["width"], (y + 0.5) * params["height"] - 0.25 * params["box"]),
        ((x + 0.5) * params["width"] + 0.25 * params["box"], (y + 0.5) * params["height"]),
    ], fill=cols[0], width=params["stroke"], outline="black")
    img.polygon([
        ((x + 0.5) * params["width"] - 0.25 * params["box"], (y + 0.5) * params["height"]),
        ((x + 0.5) * params["width"], (y + 0.5) * params["height"] + 0.25 * params["box"]),
        ((x + 0.5) * params["width"] + 0.25 * params["box"], (y + 0.5) * params["height"]),
    ], fill=cols[1], width=params["stroke"], outline="black")


def draw_deoxyhexose(img, x, y, name, **params):
    """
    Draw a filled triangle for an assigned glycan.

    Args:
        img (Draw): ImageDraw object to use for drawing.
        x (float): x coordinate of the shape on the image before offsetting
        y (float): y coordinate of the shape on the image before offsetting
        name (str): name of the glycan to select the correct color
        **params: See description in Glycan.create_snfg_image
    """
    img.polygon([
        ((x + 0.5) * params["width"] - 0.25 * params["box"], (y + 0.5) * params["height"] + 0.25 * params["box"]),
        ((x + 0.5) * params["width"], (y + 0.5) * params["height"] - 0.25 * params["box"]),
        ((x + 0.5) * params["width"] + 0.25 * params["box"], (y + 0.5) * params["height"] + 0.25 * params["box"]),
    ], fill=colors["Deoxyhexose"][name], width=params["stroke"], outline="black")


def draw_deoxyhexnac(img, x, y, name, **params):
    """
    Draw a divided triangle for an assigned glycan.

    Args:
        img (Draw): ImageDraw object to use for drawing.
        x (float): x coordinate of the shape on the image before offsetting
        y (float): y coordinate of the shape on the image before offsetting
        name (str): name of the glycan to select the correct color
        **params: See description in Glycan.create_snfg_image
    """
    img.polygon([
        ((x + 0.5) * params["width"] - 0.25 * params["box"], (y + 0.5) * params["height"] + 0.25 * params["box"]),
        ((x + 0.5) * params["width"], (y + 0.5) * params["height"] - 0.25 * params["box"]),
        ((x + 0.5) * params["width"], (y + 0.5) * params["height"] + 0.25 * params["box"]),
    ], fill="white", width=params["stroke"], outline="black")
    img.polygon([
        ((x + 0.5) * params["width"], (y + 0.5) * params["height"] + 0.25 * params["box"]),
        ((x + 0.5) * params["width"], (y + 0.5) * params["height"] - 0.25 * params["box"]),
        ((x + 0.5) * params["width"] + 0.25 * params["box"], (y + 0.5) * params["height"] + 0.25 * params["box"]),
    ], fill=colors["DeoxyhexNAc"][name], width=params["stroke"], outline="black")


def draw_di_deoxyhexose(img, x, y, name, **params):
    """
    Draw a flat rectangle for an assigned glycan.

    Args:
        img (Draw): ImageDraw object to use for drawing.
        x (float): x coordinate of the shape on the image before offsetting
        y (float): y coordinate of the shape on the image before offsetting
        name (str): name of the glycan to select the correct color
        **params: See description in Glycan.create_snfg_image
    """
    img.rectangle([
        ((x + 0.5) * params["width"] - 0.25 * params["box"], (y + 0.5) * params["height"] - 0.125 * params["box"]),
        ((x + 0.5) * params["width"] + 0.25 * params["box"], (y + 0.5) * params["height"] + 0.125 * params["box"]),
    ], fill=colors["Di-deoxyhexose"][name], width=params["stroke"], outline="black")


def draw_pentose(img, x, y, name, **params):
    """
    Draw a 5-star for an assigned glycan.

    Args:
        img (Draw): ImageDraw object to use for drawing.
        x (float): x coordinate of the shape on the image before offsetting
        y (float): y coordinate of the shape on the image before offsetting
        name (str): name of the glycan to select the correct color
        **params: See description in Glycan.create_snfg_image
    """
    img.polygon([
        ((x + 0.5) * params["width"], (y + 0.5) * params["height"] - 0.25 * params["box"]),

        ((x + 0.5) * params["width"] + 0.05902 * params["box"], (y + 0.5) * params["height"] - 0.06836 * params["box"]),
        ((x + 0.5) * params["width"] + 0.25 * params["box"], (y + 0.5) * params["height"] - 0.06836 * params["box"]),

        ((x + 0.5) * params["width"] + 0.0955 * params["box"], (y + 0.5) * params["height"] + 0.0439 * params["box"]),
        ((x + 0.5) * params["width"] + 0.16877 * params["box"], (y + 0.5) * params["height"] + 0.25 * params["box"]),

        ((x + 0.5) * params["width"], (y + 0.5) * params["height"] + 0.11328 * params["box"]),

        ((x + 0.5) * params["width"] - 0.16877 * params["box"], (y + 0.5) * params["height"] + 0.25 * params["box"]),
        ((x + 0.5) * params["width"] - 0.0955 * params["box"], (y + 0.5) * params["height"] + 0.0439 * params["box"]),

        ((x + 0.5) * params["width"] - 0.25 * params["box"], (y + 0.5) * params["height"] - 0.06836 * params["box"]),
        ((x + 0.5) * params["width"] - 0.05902 * params["box"], (y + 0.5) * params["height"] - 0.06836 * params["box"]),
    ], fill=colors["Pentose"][name], width=params["stroke"], outline="black")


def draw_3_deoxy_nonulosonic_acids(img, x, y, name, **params):
    """
    Draw a filled diamond for an assigned glycan.

    Args:
        img (Draw): ImageDraw object to use for drawing.
        x (float): x coordinate of the shape on the image before offsetting
        y (float): y coordinate of the shape on the image before offsetting
        name (str): name of the glycan to select the correct color
        **params: See description in Glycan.create_snfg_image
    """
    img.polygon([
        ((x + 0.5) * params["width"] - 0.25 * params["box"], (y + 0.5) * params["height"]),
        ((x + 0.5) * params["width"], (y + 0.5) * params["height"] - 0.25 * params["box"]),
        ((x + 0.5) * params["width"] + 0.25 * params["box"], (y + 0.5) * params["height"]),
        ((x + 0.5) * params["width"], (y + 0.5) * params["height"] + 0.25 * params["box"]),
    ], fill=colors["3-dideoxy-nunolusonic acids"][name], width=params["stroke"], outline="black")


def draw_3_9_dideocy_nonulosonic_acids(img, x, y, name, **params):
    """
    Draw a diamond for an assigned glycan.

    Args:
        img (Draw): ImageDraw object to use for drawing.
        x (float): x coordinate of the shape on the image before offsetting
        y (float): y coordinate of the shape on the image before offsetting
        name (str): name of the glycan to select the correct color
        **params: See description in Glycan.create_snfg_image
    """
    img.polygon([
        ((x + 0.5) * params["width"] - 0.25 * params["box"], (y + 0.5) * params["height"]),
        ((x + 0.5) * params["width"], (y + 0.5) * params["height"] - 0.125 * params["box"]),
        ((x + 0.5) * params["width"] + 0.25 * params["box"], (y + 0.5) * params["height"]),
        ((x + 0.5) * params["width"], (y + 0.5) * params["height"] + 0.125 * params["box"]),
    ], fill=colors["3,9-dideoxy-nunolusonic acids"][name], width=params["stroke"], outline="black")


def draw_unknown(img, x, y, name, **params):
    """
    Draw a flat hexagon for an assigned glycan.

    Args:
        img (Draw): ImageDraw object to use for drawing.
        x (float): x coordinate of the shape on the image before offsetting
        y (float): y coordinate of the shape on the image before offsetting
        name (str): name of the glycan to select the correct color
        **params: See description in Glycan.create_snfg_image
    """
    img.polygon([
        ((x + 0.5) * params["width"] - 0.25 * params["box"], (y + 0.5) * params["height"]),
        ((x + 0.5) * params["width"] - 0.125 * params["box"], (y + 0.5) * params["height"] - 0.125 * params["box"]),
        ((x + 0.5) * params["width"] + 0.125 * params["box"], (y + 0.5) * params["height"] - 0.125 * params["box"]),
        ((x + 0.5) * params["width"] + 0.25 * params["box"], (y + 0.5) * params["height"]),
        ((x + 0.5) * params["width"] + 0.125 * params["box"], (y + 0.5) * params["height"] + 0.125 * params["box"]),
        ((x + 0.5) * params["width"] - 0.125 * params["box"], (y + 0.5) * params["height"] + 0.125 * params["box"]),
    ], fill=colors["Unknown"][name], width=params["stroke"], outline="black")


def draw_assigned(img, x, y, name, **params):
    """
    Draw a pentagon for an assigned glycan.

    Args:
        img (Draw): ImageDraw object to use for drawing.
        x (float): x coordinate of the shape on the image before offsetting
        y (float): y coordinate of the shape on the image before offsetting
        name (str): name of the glycan to select the correct color
        **params: See description in Glycan.create_snfg_image
    """
    img.polygon([
        ((x + 0.5) * params["width"], (y + 0.5) * params["height"] - 0.25 * params["box"]),
        ((x + 0.5) * params["width"] + 0.25 * params["box"], (y + 0.5) * params["height"] - 0.06836 * params["box"]),
        ((x + 0.5) * params["width"] + 0.16877 * params["box"], (y + 0.5) * params["height"] + 0.25 * params["box"]),
        ((x + 0.5) * params["width"] - 0.16877 * params["box"], (y + 0.5) * params["height"] + 0.25 * params["box"]),
        ((x + 0.5) * params["width"] - 0.25 * params["box"], (y + 0.5) * params["height"] - 0.06836 * params["box"]),
    ], fill=colors["Assigned"][name], width=params["stroke"], outline="black")


def draw_other(img, x, y, **params):
    """
    Draw an empty pentagon for an unknown glycan.

    Args:
        img (Draw): ImageDraw object to use for drawing.
        x (float): x coordinate of the shape on the image before offsetting
        y (float): y coordinate of the shape on the image before offsetting
        **params: See description in Glycan.create_snfg_image
    """
    img.polygon([
        ((x + 0.5) * params["width"], (y + 0.5) * params["height"] - 0.25 * params["box"]),
        ((x + 0.5) * params["width"] + 0.25 * params["box"], (y + 0.5) * params["height"] - 0.06836 * params["box"]),
        ((x + 0.5) * params["width"] + 0.16877 * params["box"], (y + 0.5) * params["height"] + 0.75 * params["box"]),
        ((x + 0.5) * params["width"] - 0.16877 * params["box"], (y + 0.5) * params["height"] + 0.25 * params["box"]),
        ((x + 0.5) * params["width"] - 0.25 * params["box"], (y + 0.5) * params["height"] - 0.06836 * params["box"]),
    ], fill="white", width=params["stroke"], outline="black")



import os
import sys
import logging

from joblib import Parallel, delayed, parallel_backend

from glyles.glycans.utils import ParseError
from glyles.glycans.poly.glycan import Glycan


def preprocess_glycans(glycan, glycan_list, glycan_file):
    """
    Preprocess the static inputs for the parsing into one single list

    Args:
        glycan (str): single glycan to parse
        glycan_list (List[str]): list of glycans to parse
        glycan_file (str): filepath of file to read glycans from

    Returns:
        list of glycans in the order they are handed to the function, i.e. glycan, glycan_list, glycan_file
    """
    glycans = []

    # fill a list with all glycans to convert
    if glycan is not None:
        glycans.append(glycan)
    if glycan_list is not None:
        glycans += glycan_list
    if glycan_file is not None:
        # check if the file is valid and read it out
        if not os.path.isfile(glycan_file):
            raise ValueError(f"{glycan_file} does not exists, cannot read glycans.")
        for line in open(glycan_file, "r").readlines():
            glycans.append(line.strip())
    return glycans


def convert(
        glycan=None,
        glycan_list=None,
        glycan_file=None,
        glycan_generator=None,
        output_file=None,
        returning=True,
        verbose=logging.INFO,
        cpu_count=1,
        full=True,
):
    """
    Convert glycans of different input formats. All glycans have to be in IUPAC-condensed notation, but how they're
    organized, can change. Either as single glycan, list or tuples of glycans, a file, or a generator. All will be
    converted to SMILES.

    Args:
        glycan (str): Single glycan to be converted from IUPAC to SMILES
        glycan_list (List[str]): list of glycans to convert
        glycan_file (str): File to read the glycans from
        glycan_generator (generator): generator yielding iupac representation.
            Together with output_generator=True this does not create any lists
        output_file (str): File to save the converted glycans in
        returning (bool): Flag indicating to return a list of tuples
        verbose (Union[int, None]): Flag indicating to have no prints from this method
        cpu_count (int): Number of CPU cores to use for parallel processing. Behavior as described in joblib:
            https://joblib.readthedocs.io/en/latest/generated/joblib.Parallel.html
        full (bool): Flag indicating that only fully convertible glycans should be returned, i.e. all modifications
            such as 3-Anhydro-[...] are also present in the SMILES

    Returns:
        List of type (IUPAC, SMILES) items giving the converted SMILES formulas. Only if returning=True is set.
    """

    if verbose is None:
        logger = logging.getLogger()
        logger.disabled = True
    else:
        logging.basicConfig(level=verbose)

    # collect all data and return if no data were provided
    glycans = preprocess_glycans(glycan, glycan_list, glycan_file)
    if len(glycans) == 0 and glycan_generator is None:
        logging.info("List of glycans is empty")
        return

    # determine the output format
    if output_file is not None:
        if os.path.isdir(os.path.dirname(os.path.abspath(output_file))):
            output = open(output_file, "w")
        else:
            logging.warning("Path of output-file does not exist! Results will be printed on stdout.")
            output = sys.stdout
        returning = False
    else:
        if returning:
            output = []
        else:
            logging.warning("No output-file specified, results will be printed on stdout.")
            output = sys.stdout

    # convert the IUPAC strings into SMILES strings from the input list and from the input generator
    container = []
    if len(glycans) != 0:
        container.append(glycans)
    if glycan_generator is not None:
        container.append(glycan_generator)

    for container in container:
        with parallel_backend('multiprocessing', n_jobs=cpu_count):
            results = Parallel()(delayed(generate)(iupac, full) for iupac in container)
            if returning:
                output += results
            else:
                for iupac, smiles in results:
                    print(iupac, smiles, file=output, sep=",")

    if returning:
        return output
    elif output_file is not None:
        output.close()

    if verbose is None:
        logger.disabled = False


def convert_generator(
        glycan=None,
        glycan_list=None,
        glycan_file=None,
        glycan_generator=None,
        verbose=logging.INFO,
        cpu_count=1,
        full=True,
):
    """
    Convert glycans of different input formats. All glycans have to be in IUPAC-condensed notation, but how they're
    organized, can change. Either as single glycan, list or tuples of glycans, a file, or a generator. All will be
    converted to SMILES, output by a generator.

    Args:
        glycan (str): Single glycan to be converted from IUPAC to SMILES
        glycan_list (List[str]): list of glycans to convert
        glycan_file (str): File to read the glycans from
        glycan_generator (generator): generator yielding iupac representation.
            Together with output_generator=True this does not create any lists
        verbose (Union[int, None]): Flag indicating to have no output-messages from this method
        cpu_count (int): Number of CPU cores to use for parallel processing. Behavior as described in joblib:
            https://joblib.readthedocs.io/en/latest/generated/joblib.Parallel.html
        full (bool): Flag indicating that only fully convertible glycans should be returned, i.e. all modifications
            such as 3-Anhydro-[...] are also present in the SMILES

    Returns:
        Generator generating pairs of type (IUPAC, SMILES) items giving the converted SMILES formulas for the IUPACs.
    """
    if verbose is None:
        logger = logging.getLogger()
        logger.disabled = True
    else:
        logging.basicConfig(level=verbose)

    glycans = preprocess_glycans(glycan, glycan_list, glycan_file)
    if len(glycans) == 0 and glycan_generator is None:
        logging.info("List of glycans is empty")
        return

    # Convert the glycans ...
    if len(glycans) != 0:
        for glycan in glycans:
            yield generate(glycan, full)

    # Convert the glycans ...
    if glycan_generator is not None:
        for glycan in glycan_generator:
            yield generate(glycan, full)

    if verbose is None:
        logger.disabled = False


def generate(glycan, full):
    """
    Actually generate the SMILES string based on the glycan given in IUPAC notation.

    Parameters:
        glycan (str): Glycan molecule described by its IUPAC string
        full (bool): flag indicating to only return SMILES string that include all modifications from the IUPAC

    Returns:
        A pair of glycan represented with its IUPAC string and SMILES string
    """
    try:
        # ... by passing them to the glycan class to parse them and return them as intended
        return glycan, Glycan(glycan, full=full).get_smiles()

    # catch any exception at glycan level to not destroy the whole pipeline because of one mis-formed glycan
    except ParseError as e:
        logging.error(f"A parsing error occurred with {glycan}: {e.__class__}\n"
                      f"Error message: {e.__str__()}")
        return glycan, ""
    except Exception as e:
        logging.error(f"An unexpected exception occurred with with {glycan}. This glycan cannot be parsed."
                      f"Error message: {e.__str__()}")
        return glycan, ""



import copy
import re

import numpy as np

from glyles.glycans.utils import sanitize_smiles


class Merger:
    """
    Merge the tree of monomers into a SMILES representation of the complete molecule.
    """

    def __init__(self, factory):
        """
        Create a merger class to merge a parsed glycan tree into a SMILES string.

        Args:
            factory (MonomerFactory): factory instance to use to generate the monomers for the glycan tree from
        """
        self.factory = factory

    def merge(self, t, root_orientation="n", start=100):
        """
        Merge the provided tree of monomers enriched with the glycans in the nodes and information on the bindings
        between two monomer-nodes in the edges. The input graph is not changed during this process, as the methods
        deep-copies the glycan representing datastructure.

        Args:
            t (networkx.DiGraph): Graph representing the glycan to compute the whole SMILES representation for.
            root_orientation (str): Orientation of the root monomer in glycan (can be a (= alpha) or b (= beta))
            start (int): index of the atom to start the SMILES generation from

        Returns:
            SMILES representation as string
        """
        # deep-copy the tree to not modify the provided argument
        t = copy.deepcopy(t)

        # first mark the atoms that will be replaced in a binding of two monomers
        self.__mark(t, 0, f"({root_orientation}1-?)")

        # return the string that can be computed from connecting the monomers as marked above
        if np.where(t.nodes[0]["type"].get_features()[:, 1] == start)[0].size != 0:
            position = int(np.argwhere(t.nodes[0]["type"].get_features()[:, 1] == start).squeeze())
        else:
            position = int(np.argwhere(t.nodes[0]["type"].get_features()[:, 1] == 1).squeeze())
        return self.__merge(t, 0, position, 0)

    def __mark(self, t, node, p_edge):
        """
        Recursively mark in every node of the molecule which atoms are being replaced by bound monomers.

        Args:
            t (networkx.DiGraph): Tree representing the glycan to compute the whole SMILES representation for.
            node (int): ID of the node to work on in this method
            p_edge (str): edge annotation to parent monomer

        Returns:
            Nothing
        """
        # get children nodes
        children = [x[1] for x in t.edges(node)]

        # set chirality of atom binding parent
        if p_edge is not None and t.nodes[node]["type"].is_non_chiral():
            t.nodes[node]["type"] = t.nodes[node]["type"].to_chirality(p_edge[1], self.factory)

        # check for validity of the tree, ie if it's a leaf (return, nothing to do) or has too many children (Error)
        if len(children) == 0:  # leaf
            return
        if len(children) > 4:  # too many children
            raise NotImplementedError("Glycans with maximal branching factor greater then 3 not implemented.")

        # iterate over the children and the atoms used to mark binding atoms in my structure
        for child, atom in zip(children, t.nodes[node]["type"].get_dummy_atoms()):
            binding = re.findall(r'\d+', t.get_edge_data(node, child)["type"])[1]

            t.nodes[node]["type"].mark(int(binding), *atom)
            self.__mark(t, child, t.get_edge_data(node, child)["type"])

    def __merge(self, t, node, start, ring_index):
        """
        Recursively merge every node of the molecule with its children and get the SMILES representation of the
        subtree.

        Args:
            t (networkx.DiGraph): Graph representing the glycan to compute the whole SMILES representation for.
            node (int): ID of the node to work on in this method
            start (int): ID of the atom in the inner graph to start from when generating the SMILES string
            ring_index (int): Index of the ring to use when generating the SMILES strings

        Returns:
            SMILES representation of the subtree of the given node
        """
        # get my children and compute my SMILES string
        children = [x[1] for x in t.edges(node)]
        me = t.nodes[node]["type"].to_smiles(ring_index, root_id=start)

        # check for validity of the tree, ie if it's a leaf
        if len(children) == 0:  # leaf
            return me

        # iterate over the children and the atoms used to mark binding atoms
        for child, (o_atom, n_atom) in zip(children, t.nodes[node]["type"].get_dummy_atoms()):
            binding = re.findall(r'\d+', t.get_edge_data(node, child)["type"])[0]

            child_start = t.nodes[child]["type"].root_atom_id(int(binding))
            if child_start == -1:
                raise ValueError("No child start found.")

            # get the SMILES of this child and plug it in the current own SMILES
            child_smiles = self.__merge(t, child, child_start, ring_index + 1)
            if o_atom[1] in me:
                me = re.sub(o_atom[2], child_smiles, me)
            elif n_atom[1] in me:
                me = re.sub(n_atom[2], "N(" + child_smiles[1:] + ")", me)
            me = sanitize_smiles(me)
            # me = me.replace("((", "(").replace("))", ")")
            # me = me.replace(atom, child_smiles)
        return me
